#!/bin/bash
# Note: There are a few bashisms in this script.
#       (here string, substring expansion, shorthand stderr redirection, process substitution)
################################################################################
# environment variables used:
#   JJPEDIR = location of JJPE files
#   GAMEDIR = location of game files
################################################################################
if [ -z "$JJPEDIR" ] || [ -z "$GAMEDIR" ]
then
  printf "Environment variable(s) not set. Check JJPEDIR, GAMEDIR\n"
  exit 2
fi

################################################################################
# globals
################################################################################
TYPE=""
temp_partition="/jjpe/temp"
jjpe_dest="${temp_partition}/destjjpe"
perm_dest="${temp_partition}/destperm"
update_path=""
dest_device="/dev/sda"
cur_part=""
new_part=""
update_type_file=""
update_type=""
# needed for partclone
export TERM=linux
PART_EFI=""
PART_BOOT=""
PART_ROOTA=""
PART_ROOTB=""
PART_PERMA=""
PART_PERMB=""
PART_TEMP=""
# get the UUIDs for the file systems
. /jjpe/gen1/scripts/fs_uuids.sh

################################################################################
# functions
################################################################################
function find_dest_device {
  # this logic is copied from jjp-ocs-sr
  local dest_disk="sda"
  local part_prefix=""
  while read -r line
  do
    read -a disk_info <<< $line
    if [ "${disk_info[0]}" = "disk" ] && [ "${disk_info[1]}" = "0" ]
	then
      # found a removable disk
      case "${disk_info[2]}" in
      sd*)		# SATA disk
        dest_disk="${disk_info[2]}"
        part_prefix=""
        break
        ;;
      nvme*)	# NVMe disk - need partition prefix
        dest_disk="${disk_info[2]}"
        part_prefix="p"
        break
        ;;
      esac
    fi
  done < <( lsblk -dln -e 7 -o TYPE,RM,NAME -x NAME )

  dest_device="/dev/${dest_disk}"
  # these are the device names of each partition
  PART_EFI="${dest_device}${part_prefix}1"
  PART_BOOT="${dest_device}${part_prefix}2"
  PART_ROOTA="${dest_device}${part_prefix}3"
  PART_ROOTB="${dest_device}${part_prefix}5"
  PART_PERMA="${dest_device}${part_prefix}4"
  PART_PERMB="${dest_device}${part_prefix}6"
  PART_TEMP="${dest_device}${part_prefix}7"
}

################################################################################
# TODO: remove all the fstab writing stuff and switch to our custom mount generator
write_fstab_a_file() {
if grep -q 'JJPE' "${jjpe_dest}/etc/fstab"
then
	# new fstab file - don't overwrite
	return
fi
cat << FSTAB_A > "${jjpe_dest}/etc/fstab"
UUID=${FS_UUID_ROOTA} /           ext4  errors=remount-ro,discard,noatime  0  1
UUID=${FS_UUID_BOOT}  /boot       ext4  errors=remount-ro,discard,noatime  0  1
UUID=${FS_UUID_PERMA} /jjpe/perm  ext4  errors=remount-ro,discard,noatime  0  2
UUID=${FS_UUID_TEMP}  /jjpe/temp  ext4  errors=remount-ro,discard,noatime  0  2
UUID=${FS_UUID_EFI}   /boot/efi   vfat  umask=0077,discard,noatime,noauto,x-systemd.automount,x-systemd.idle-timeout=10s  0  1
FSTAB_A
}

################################################################################
write_fstab_b_file() {
if grep -q 'JJPE' "${jjpe_dest}/etc/fstab"
then
	# new fstab file - don't overwrite
	return
fi
cat << FSTAB_B > "${jjpe_dest}/etc/fstab"
UUID=${FS_UUID_ROOTB} /           ext4  errors=remount-ro,discard,noatime  0  1
UUID=${FS_UUID_BOOT}  /boot       ext4  errors=remount-ro,discard,noatime  0  1
UUID=${FS_UUID_PERMB} /jjpe/perm  ext4  errors=remount-ro,discard,noatime  0  2
UUID=${FS_UUID_TEMP}  /jjpe/temp  ext4  errors=remount-ro,discard,noatime  0  2
UUID=${FS_UUID_EFI}   /boot/efi   vfat  umask=0077,discard,noatime,noauto,x-systemd.automount,x-systemd.idle-timeout=10s  0  1
FSTAB_B
}

################################################################################
write_fstab_file() {
  if [ "$cur_part" = 'a' ]
  then
    write_fstab_b_file
  elif [ "$cur_part" = 'b' ]
  then
    write_fstab_a_file
  fi
}

################################################################################
function delete_update_files {
  rm -rf "$update_path"
  rm -f "${temp_partition}"/*.zip*
}

################################################################################
function force_img_next_update {
  delete_update_files
  touch /boot/grub/forceimg
  echo 'rsync failed to copy' > "${temp_partition}/rprogress"
  echo 'partclone failed to restore' > "${temp_partition}/pcprogress"
}

################################################################################
function swap_grub {
  /boot/grub/swapgrub.sh -p "$new_part"
  echo "$new_part" > /boot/grub/curgrub
}

################################################################################
function mount_other_parts {
  mkdir -p "$jjpe_dest"
  mkdir -p "$perm_dest"
  if [ "$cur_part" = 'a' ]
  then
    mount "$PART_ROOTB" "$jjpe_dest"
    mount "$PART_PERMB" "$perm_dest"
    new_part='b'
  elif [ "$cur_part" = 'b' ]
  then
    mount "$PART_ROOTA" "$jjpe_dest"
    mount "$PART_PERMA" "$perm_dest"
    new_part='a'
  fi
  sleep 1
}

################################################################################
function umount_other_parts {
  if [ "$cur_part" = 'a' ]
  then
    umount "$PART_ROOTB"
    umount "$PART_PERMB"
  elif [ "$cur_part" = 'b' ]
  then
    umount "$PART_ROOTA"
    umount "$PART_PERMA"
  fi
  sleep 1
  rm -rf "$jjpe_dest"
  rm -rf "$perm_dest"
}

################################################################################
function copy_perm_and_settings {
  mkdir -p "${perm_dest}/vf"
  rm -rf "${perm_dest}"/vf/*
  cp -r /jjpe/perm/vf/* "$perm_dest/vf/"
  mkdir -p "${jjpe_dest}/etc/NetworkManager/system-connections"
  rm -rf "${jjpe_dest}/etc/NetworkManager/system-connections"/*
  cp -r /etc/NetworkManager/system-connections/* \
    "${jjpe_dest}/etc/NetworkManager/system-connections/"
  cp /etc/timezone "${jjpe_dest}/etc"
  cp /etc/hostname "${jjpe_dest}/etc"
}

################################################################################
function set_file_perms {
  chmod --silent +x "${jjpe_dest}${GAMEDIR}/game"
  chmod --silent +x "${jjpe_dest}${JJPEDIR}"/*.sh
  chmod --silent +x "${jjpe_dest}${JJPEDIR}"/*.pl
  chmod --silent --recursive +x "${jjpe_dest}${JJPEDIR}/scripts/"
}

################################################################################
function update_net_delta {
  mount_other_parts
  if rsync -ca --delay-updates --perms --info=progress2 --no-i-r --stats \
    --exclude='ui.dat' --exclude='update_type' \
    --log-file="${temp_partition}/update.log" "${update_path}/" \
    "${jjpe_dest}${JJPEDIR}/" &> "${temp_partition}/rprogress"
  then
    swap_grub
    copy_perm_and_settings
    set_file_perms
    sync
    echo 'rsync finished successfully' > "${temp_partition}/rprogress"
  else
    force_img_next_update
    echo 'rsync failed to copy' > "${temp_partition}/rprogress"
  fi
  umount_other_parts
}

################################################################################
function update_efi_img {
  local target="$PART_EFI"
  local uuid="$FS_UUID_EFI"
  kill -9 $(lsof -t $target 2> /dev/null)
  sleep 1
  umount "$target"
  sleep 1

  set -o pipefail
  cat "${update_path}/img"/sda1.vfat-ptcl-img.gz.a* | \
    gunzip -c | partclone.vfat -F -I -C -R -L "${temp_partition}/efi_img.log" \
    -r -s - -o "$target" &> /dev/null
  if [ $? -eq 0 ]
  then
    printf "\x${uuid:7:2}\x${uuid:5:2}\x${uuid:2:2}\x${uuid:0:2}" \
    | dd bs=1 seek=67 count=4 conv=notrunc of="$target"
  else
    force_img_next_update
    echo 'partclone failed to restore' > "${temp_partition}/pcprogress"
  fi
}

################################################################################
function update_boot_img {
  local target="$PART_BOOT"
  local uuid="$FS_UUID_BOOT"
  kill -9 $(lsof -t $target 2> /dev/null)
  sleep 1
  umount "$target"
  sleep 1

  set -o pipefail
  cat "${update_path}/img"/sda2.ext4-ptcl-img.gz.a* | \
    gunzip -c | partclone.ext4 -F -I -C -R -L "${temp_partition}/boot_img.log" \
    -r -s - -o "$target" &> /dev/null
  if [ $? -eq 0 ]
  then
    e2fsck -f -y "$target" &> /dev/null
    tune2fs "$target" -f -U "$uuid" > /dev/null 2>&1
    mount "$target" /boot
    sleep 1	
  else
    force_img_next_update
    echo 'partclone failed to restore' > "${temp_partition}/pcprogress"
  fi
}

################################################################################
function update_net_img {
  if [ "$cur_part" = 'a' ]
  then
    local target="$PART_ROOTB"
    local uuid="$FS_UUID_ROOTB"
  elif [ "$cur_part" = 'b' ]
  then
    local target="$PART_ROOTA"
    local uuid="$FS_UUID_ROOTA"
  fi
  set -o pipefail
  cat "${update_path}/img"/sda3.ext4-ptcl-img.gz.a* | \
    gunzip -c | partclone.ext4 -F -I -C -R -L "${temp_partition}/update.log" \
    -r -s - -o "$target" &> "${temp_partition}/pcprogress"
  if [ $? -eq 0 ]
  then
    e2fsck -f -y "$target" &> /dev/null
    resize2fs "$target" &> /dev/null
    e2fsck -f -y "$target" &> /dev/null
    tune2fs "$target" -f -U "$uuid" > /dev/null 2>&1
    mount_other_parts
    write_fstab_file
    swap_grub
    copy_perm_and_settings
    set_file_perms
    umount_other_parts
    rm -f /boot/grub/forceimg
    sync
    echo 'partclone successfully restored' > "${temp_partition}/pcprogress"
  else
    force_img_next_update
    echo 'partclone failed to restore' > "${temp_partition}/pcprogress"
  fi
}

################################################################################
function update_net {
  if [ "$update_type" = 'delta' ]
  then
    update_net_delta
  elif [ "$update_type" = 'img' ]
  then
    update_efi_img
    update_boot_img
    update_net_img
  else
    force_img_next_update
  fi
}

################################################################################
function update_usb_delta {
  mount_other_parts
  if rsync -ca --delay-updates --perms --info=progress2 --no-i-r --stats \
    --exclude='ui.dat' --exclude='update_type' \
    --log-file="${temp_partition}/update.log" "${update_path}/" \
    "${jjpe_dest}${JJPEDIR}/" &> "${temp_partition}/rprogress"
  then
    swap_grub
    copy_perm_and_settings
    set_file_perms
    sync
	# TODO: the umount should be moved to copyusbdeltafiles.sh
    umount -f /mnt/usb &> /dev/null
    sleep 1
    echo 'rsync finished successfully' > "${temp_partition}/rprogress"
  else
    force_img_next_update
    echo 'rsync failed to copy' > "${temp_partition}/rprogress"
  fi

  umount_other_parts
}

################################################################################
function sync_net_delta {
  mount_other_parts
  if rsync -ca --delay-updates --perms --info=progress2 --no-i-r --stats \
    --exclude='ui.dat' --exclude='update_type' \
    --log-file="${temp_partition}/syncotherpart.log" "${update_path}/" \
    "${jjpe_dest}${JJPEDIR}/" &> "${temp_partition}/rprogress"
  then
    set_file_perms
    delete_update_files
    sync
    echo 'rsync finished successfully' > "${temp_partition}/rprogress";
  else
    force_img_next_update
    echo 'rsync failed to copy' > "${temp_partition}/rprogress"
  fi

  umount_other_parts
}

################################################################################
function sync_net_img {
  if [ "$cur_part" = 'a' ]
  then
    local target="$PART_ROOTB"
    local uuid="$FS_UUID_ROOTB"
  elif [ "$cur_part" = 'b' ]
  then
    local target="$PART_ROOTA"
    local uuid="$FS_UUID_ROOTA"
  fi
  set -o pipefail
  cat "${update_path}/img"/sda3.ext4-ptcl-img.gz.a* | \
    gunzip -c | partclone.ext4 -F -I -C -R -L \
    "${temp_partition}/syncotherpart.log" \
    -r -s - -o "$target" &> "${temp_partition}/pcprogress"
  if [ $? -eq 0 ]
  then
    e2fsck -f -y "$target" &> /dev/null
    resize2fs "$target" &> /dev/null
    e2fsck -f -y "$target" &> /dev/null
    tune2fs "$target" -f -U "$uuid" > /dev/null 2>&1
    mount_other_parts
    write_fstab_file
    set_file_perms
    umount_other_parts
    delete_update_files
    sync
    echo 'partclone successfully restored' > "${temp_partition}/pcprogress"
  else
    force_img_next_update
    echo 'partclone failed to restore' > "${temp_partition}/pcprogress"
  fi
}

################################################################################
function sync_usb_delta {
  mount_other_parts
  if rsync -ca --delay-updates --perms --info=progress2 --no-i-r --stats \
    --exclude='ui.dat' --exclude='update_type' \
    --log-file="$temp_partition/syncotherpart.log" "${update_path}/" \
    "${jjpe_dest}$JJPEDIR/" &> "${temp_partition}/rprogress"
  then
    set_file_perms
    delete_update_files
    sync
    echo 'rsync finished successfully' > "${temp_partition}/rprogress";
  else
    force_img_next_update
    echo 'rsync failed to copy' > "${temp_partition}/rprogress"
  fi
  umount_other_parts
}

################################################################################
function sync_old_partition {
  rm -f "$update_type_file"
  if [ "$update_type" = 'delta' ]
  then
    sync_net_delta
  elif [ "$update_type" = 'img' ]
  then
    sync_net_img
  elif [ "$update_type" = 'usb' ]
  then
    sync_usb_delta
  else
    force_img_next_update
  fi
}

################################################################################
function show_usage {
cat << USAGE
OPTIONS:
 -t, type  net, usb, or sync
 -?,       show this message and exit
 update_dir
USAGE
}

################################################################################
# process command line args
################################################################################
# put a : in the beginning of the option string to suppress getopt error
# message. follow an option with a : to make getopts require an argument.
# it can be accessed in the variable OPTARG.
while getopts ":t:n:?" opt; do
  case $opt in
    t) TYPE="$OPTARG" ;;
   \?) show_usage
       exit 1
  esac
done
shift $(($OPTIND - 1))

if [ $# -ge 1 ]
then
  updatedir=$1
else
  printf "missing arg: updatedir not specified\n"
  exit 2
fi
update_path="${temp_partition}/${updatedir}"
update_type_file="${update_path}/update_type"
preupdate_script=${update_path}/preupdate.sh

################################################################################
# main
################################################################################
sync
sleep 2
if [ -f "$update_type_file" ]
then
  find_dest_device
  update_type=$(<"$update_type_file")
  cur_part=$(</boot/grub/curgrub)  
  if [ -e "$preupdate_script" ]
  then
    # run the pre-update script found in the update package
    /bin/dash "$preupdate_script" "$dest_device" "$TYPE" "$update_type"
    if [ $? -eq 100 ]
    then
      # return value of 100 means that the pre-update script did all the work
      # bail out
      exit 0
    fi
  fi
  if [ "$TYPE" = 'net' ]
  then
    update_net
  elif [ "$TYPE" = 'usb' ]
  then
    update_usb_delta
  elif [ "$TYPE" = 'sync' ]
  then
    sync_old_partition
  else
    force_img_next_update
  fi
else
  force_img_next_update
fi
